import sys
import logging
from tlx.util import paginate
from tlx.dynamodb.batch import batch_delete, get_ddb_table
logger = logging.getLogger()
logger.setLevel(logging.INFO)

# Bellow thows a syntax error in python 2.7. no idea why
# py_version = float(f"{sys.version_info.major}.{sys.version_info.minor}")
py_version = float("{}.{}".format(sys.version_info.major, sys.version_info.minor))

if py_version < 3.6:
    raise RuntimeError("This module is not for python <3.6.")


def add_key(table, key, item):
    """ If the item doesn't exist yet. Create the key.
        !! Does NOT support tables with Partition and Sort keys Yet.
    """

    key_names = [k for k in key]
    full_item = {**key, **item}  # noqa: E999   - only invalid in old pythons
    logger.debug(f'submitting item: {full_item}')  # noqa: E999   - only invalid in old pythons

    logger.info(f"Attempting to add new record for: {key} ")
    res = table.put_item(
        Item=full_item,
        ConditionExpression=f"attribute_not_exists({key_names[0]})",  # TODO fix for items with Partition and Sort key
    )

    logger.info(f"Successfully added new record.")
    logger.debug(f"{full_item}")
    return res['ResponseMetadata']['HTTPStatusCode']


def append_to_list_field(table, key, field_to_update, expression_attribute_names, new_item, add_missing_key=False):
    """ Appends `new_item` to existing item `field_to_update` for a given `key`.

        N.B. If the field is not found, attempts are made to create the entire path of `field_to_update`.
        If the key is not found, attempt to create it.

        Args:
            table (boto3 Table):
            key (dict): Primary key for the table. e.g {'matchid': matchid}
            field_to_update (list): e.g ['providers', '#dataproviderid'].  Variables for substitution should
                                    start with a hash and have values defined in `expression_attribute_names`.
            expression_attribute_names (dict): Expansion of variables names in the field to update.
                                               E.g {'#dataproviderid': "BILL"}
            new_item (list): To be appended to a list, the object must be a list.
            add_missing_key (bool): Default False. If True add the primary key for the table if it is not found.
                                    Otherwise returns dict of data to be added by user.

         Returns:
            int (200)                   - If successfully added item to Dynamodb
            dict {unadded data}         - If the primary key was not found. use `add_key` to add it.
    """
    try:
        str_path = '.'.join(field_to_update)
        # TODO: Consider adding a check for if key exists here rather than in the `_add_new_map_field` call stack
        # Current model is optimised for frequent updates of existing keys.  If keys are frequently being added
        # it would be more efficient to try adding the key first and then this method.
        res = table.update_item(
            Key=key,
            UpdateExpression=f"SET {str_path} = list_append(:new_item, {str_path})",
            ExpressionAttributeNames=expression_attribute_names,
            ExpressionAttributeValues={":new_item": new_item},
            ConditionExpression=f"attribute_exists({str_path})",
        )
        logger.info(f"Successfully added new odds to: {str_path}")
        logger.debug(f'{new_item}')
        return res['ResponseMetadata']['HTTPStatusCode']
    except table.meta.client.exceptions.ConditionalCheckFailedException:
        # Table missing some amount of structure
        logger.info(f"field not found: {str_path}")

    return add_new_map_field(table, key, field_to_update[:-1], field_to_update[-1], expression_attribute_names, new_item, add_missing_key)


def add_new_map_field(table, key, path, field_to_add, expression_attribute_names, data, add_missing_key=False, replace_existing=False):
    """ Adds a new field (field_to_add) to an existing path in a map (path) for a given Primary key (key).

        N.B If the path doesn't exist, the function attempts to create that path.  If the key doesn't exist,
        it will attempt to create IFF `add_missing_key=True`.

        Args:
            table (boto3 Table):
            key (dict): Must match the schema for `table`,
            path (list): List of field name strings in the map path. Use # if the field should be substitued for a value.
                         Specify that value in `expression_attribute_names`. E.g ['root', '#firstkey', '2ndkey']
            field_to_add (string): The name of the new field to add at the end of existing map described by `path`.
            expression_attribute_names (dict): Substitutes for the #values in `path`. E.g {'#firstkey': 'TeamMembers'}
            data (any): Value to be saved under `path`+`field_to_add`.  Must be valid datatypes for Dynamodb.
            add_missing_key (bool): Default False. If True add the primary key for the table if it is not found.
                                    Otherwise returns dict of data to be added by user.
            replace_existing (bool): Default False. If True, the `field_to_add` may exist already.  In the case that
                                     it does, it is replaced with `data`. If False, an error will be returned.
         Returns:
            int (200)                   - If successfully added item to Dynamodb
            dict {unadded data}         - If the primary key was not found. use `add_key` to add it.
    """

    if not path:  # Limit of recursion (Primary key was not found)
        return add_key(table, key, field_to_add, data) if add_missing_key else {field_to_add: data}

    str_path = '.'.join(path)
    # Default to error instead of overwrite
    condition_exp = f"attribute_not_exists({str_path}.{field_to_add}) AND attribute_exists({str_path})"
    if replace_existing:
        condition_exp = f"attribute_exists({str_path})"

    try:
        res = table.update_item(
            Key=key,
            UpdateExpression=f"SET {str_path}.{field_to_add} = :new_item",  # want dots to expand
            ExpressionAttributeNames=expression_attribute_names,
            ExpressionAttributeValues={":new_item": data},
            ConditionExpression=condition_exp,
        )
        logger.info(f"Successfully added new data to: {str_path}.{field_to_add}")
        logger.debug(f'{data}')
        return res['ResponseMetadata']['HTTPStatusCode']
    except table.meta.client.exceptions.ConditionalCheckFailedException:
        logger.info(f"Path ({str_path}) not found.")
    except table.meta.client.exceptions.ClientError as ce:
        # The document path has changed. Allow it to overwrite next itteration otherwise raise
        if ce.response['Error']['Code'] in ('ValidationException',):
            msg = "The document path has changed. Overwriting existing data!"
            logger.warn(msg)
            replace_existing = True
        else:
            raise ce

    # Must remove otherwise boto complains 'unused in expressions' next time
    new_data = {expression_attribute_names.pop(field_to_add, field_to_add): data}

    # Recursively try to add until we empty the 'path' variable
    return add_new_map_field(table, key, path[:-1], path[-1], expression_attribute_names, new_data, replace_existing=replace_existing)


def clear_table(table):
    """ in Alpha

        - WILL IMMEDIATELY DELETE ALL ITEMS WITHOUT CONFIRMATION !

        NOT TESTED with Primary and Sort Key Tables !!!
        TODO:
            - test on multikey tables
            - make cli app for it
    """

    table = get_ddb_table(table)
    ddb = table.meta.client

    table_keys = [key['AttributeName'] for key in table.key_schema]
    all_ids = ({key: r[key] for key in table_keys} for r in paginate(ddb.scan, TableName=table.name))
    batch_delete(table, all_ids)


def full_scan(table, **table_scan_params):
    """ Paginates fully over the table resource scan method wich is not natively pagable.
        Although the client scan method is pagable using with the generic paginator in tlx,
        it won't accept the FilterExpression from `boto3.dynamodb.conditions` (`Key, Attr`)

        This method takes normal table scan parameters and returns the complete list.

        e.g
            >>> from boto3.dynamodb.conditions import Key, Attr
            >>> table_scan_params = dict(
            ...     # IndexName='User_Latest_Predictions2',
            ...     # ProjectionExpression='itemId',
            ...     FilterExpression=Attr('userId').eq(USERID),
            ... )
            >>> items = tlx.dynamodb.table.full_scan(table_name, **table_scan_params)

    """

    table = get_ddb_table(table)

    items = []
    scan_incomplete = True
    while scan_incomplete:
        res = table.scan(**table_scan_params)
        items.extend(res['Items'])
        try:
            table_scan_params['ExclusiveStartKey'] = res['LastEvaluatedKey']
        except KeyError:  # Last item
            scan_incomplete = False
    return items



import string
import random
from uuid import uuid4


def string_from_datetime(dt_obj, sep=None, timespec='milliseconds'):
    """string format from python datetime"""
    # Pandas datetime is not yet implimenting `timespec`
    # https://github.com/pandas-dev/pandas/issues/26131
    try:
        return dt_obj.isoformat(sep=sep, timespec=timespec)
    except TypeError as e:
        msg = ("\n\nATTN!! If your datetime is actually "
               "`pandas._libs.tslibs.timestamps.Timestamp`"
               " then you may raise this exception.  Use `to_pydatetime` "
               "to convert it.\n\n")
        print(msg)
        raise e


def get_dynamo_compatible_uuid():
    """ Returns a 32 char random hash garenteeing lowercase letter at the begining.
        (DynamoDB has issues with digits at the start)
    """
    return random.choice(string.ascii_lowercase) + uuid4().hex[1:]



def paginate(method, **kwargs):
    """ Automatically paginates through result lists regardless of what type of marker/token/continuation
        AWS decided to use on that service. Will raise `OperationNotPageableError` if the operation is
        not pagable.

        e.g Get all Log groups rather than just the first 50
            >>> log_groups = [lg['logGroupName'] for lg in paginate(logs.describe_log_groups)]  # `nextToken`
            >>> all_resources = [ r for r in paginate(apig.get_resources, restApiId=rest_apis[0]['id']]  # `position`
            >>> all_stacks = [ s for s in paginate(cfn.list_stacks)]  # `NextToken`
            >>> all_roles = [r for r in paginate(iam.list_roles)]  # `Marker`
            >>> all_objects = [ob for ob in paginate(s3.list_objects, Bucket=bucket_name, MaxKeys=10)]
    """

    client = method.__self__
    paginator = client.get_paginator(method.__name__)
    for page in paginator.paginate(**kwargs).result_key_iters():
        try:
            for result in page:
                yield result
        except TypeError:
            pass



import os
import boto3
import logging
from getpass import getpass
from botocore.credentials import DeferredRefreshableCredentials
from botocore.exceptions import ClientError


class Session(boto3.session.Session):
    """Returns an AWS session instance.  If no parameters are provided, keys from `.aws/credentials` 'default' profile are used.

        Kwargs (all optional):
            profile (str): A profile with API keys as configured in ~/.aws/credentials file. Not to be used with other parmeters.
                If your role needs an MFA token you will be prompted for input.
            region (str): An AWS region
            role (str): AWS Arn of the role the user is assuming. If None, the users identity is used.
            mfa_serial (str): AWS Arn of the users Multi-Factor authentication device.
            mfa_token (str): 6 digit Multi-Factor Authentication code.

        Example:
            session = Session()
            s3client = session.client('s3')


        NB: If you already have AWS environment variables in the root shell they will take precedence. Override them by specifing any arguments.
    """

    def __init__(self, profile=None, region=None, role=None, mfa_serial=None, mfa_token=None):
        if profile and role:
            raise AttributeError("Either a profile should be used OR a role assumed. Not both.")

        params = {
            'region_name': region,
        }

        # 1. Allow pass through if env vars already exist - Dont need extra params boto will get them
        temp_creds_already_exist = len(
            {'AWS_SECRET_ACCESS_KEY', 'AWS_ACCESS_KEY_ID', 'AWS_SESSION_TOKEN'}.intersection(os.environ)
        ) == 3
        if temp_creds_already_exist:
            pass  # this is all we need  TODO: test on expired ones

        # 2. If assuming a role get temp creds
        if role:  # ! Not `elif` because role should override
            creds = _assume_role(role, mfa_serial, mfa_token)
            params.update({
                'aws_access_key_id': creds['AccessKeyId'],
                'aws_secret_access_key': creds['SecretAccessKey'],
                'aws_session_token': creds['SessionToken'],
            })
        # 4. If pofile is user profile and has mfa, populate env var params like role
        profile_mfa_serial = _get_mfa_serial_if_user(profile) if profile and not temp_creds_already_exist else None
        if profile_mfa_serial:
            if not mfa_token:
                mfa_token = getpass(f'Enter MFA code for {profile_mfa_serial}: ')

            user_base_session = boto3.session.Session(profile_name=profile)
            creds = user_base_session.client('sts').get_session_token(
                SerialNumber=profile_mfa_serial,
                TokenCode=mfa_token,
            )['Credentials']
            params.update({
                'aws_access_key_id': creds['AccessKeyId'],
                'aws_secret_access_key': creds['SecretAccessKey'],
                'aws_session_token': creds['SessionToken'],
            })

        # 3. If profile add it to input params
        elif profile:
            params['profile_name'] = profile

        try:
            boto3.session.Session.__init__(self, **{k: v for k, v in params.items() if v})
        except ClientError as e:
            logging.error("Do the API keys in `~/.aws/credentials` match those required? \n{}".format(e))

    def get_session_creds(self):
        creds = self.get_credentials()

        if isinstance(creds, DeferredRefreshableCredentials):
            # prompt for MFA token
            creds = creds.get_frozen_credentials()

        return creds


def _get_mfa_serial_if_user(profile):
    """Finds users mfa_serial from ~/.aws/credentials"""

    profile = profile or 'default'
    correct_profile = False
    is_user_profile = None  # Should return None if Role profile
    identified_mfa_serial = None

    with open(os.path.expanduser('~/.aws/credentials'), 'r') as f:
        for line in f:
            if line.startswith(f"[{profile}]"):
                correct_profile = True
            elif correct_profile and line.startswith('mfa_serial'):
                identified_mfa_serial = line.split('=')[-1].strip()
            elif correct_profile and line.startswith('aws_access_key_id'):
                is_user_profile = True
            elif line == '\n':
                if correct_profile:
                    break  # No need to search futher

    if not correct_profile:
        msg = f"Profile '{profile}' not found.  Typo?"
        raise Exception(msg)

    if is_user_profile:
        return identified_mfa_serial
    # Otherwise we have found a Role or other that will automatically pick
    # up the MFA by boto
    return None


def _assume_role(role, mfa_serial, mfa_token):
    params = {
        "RoleArn": role,
        "RoleSessionName": role,
        "SerialNumber": mfa_serial,
        "TokenCode": mfa_token
    }
    return boto3.client('sts').assume_role(**{k: v for k, v in params.items() if v})['Credentials']



#!/usr/bin/env python

from __future__ import print_function
import click
import os
import sys
from tlx.util import Session


@click.command(context_settings=dict(max_content_width=120))
@click.option('--profile', '-p', default='default', help="A profile defined in `~/.aws/credentials`.  If it requires an MFA token a prompt will be given")
@click.option('--quiet', default=False, is_flag=True, help='If using the outputs directly, see --help.')
def main(profile, quiet):
    """
        GET AWS CREDS (GAC):

        Gets temporary AWS session from a profile (user or role).  Allows you to export into your shell and run tools expecting the AWS standard environment variables. Namely:

        \b
            - AWS_ACCESS_KEY_ID
            - AWS_SECRET_ACCESS_KEY
            - AWS_SESSION_TOKEN

        Your aws credentials file (default: `~/.aws/credentials`) should be configured as follows:

        A Role profile would look like:

        \b
            [roleprofilename]
            region = us-east-2
            role_arn = arn:aws:iam::************:role/<IAM_Role_name>
            mfa_serial = arn:aws:iam::<AccountID>:mfa/<IAM_user_name>
            source_profile = default

        Or a user profile might look like:

        \b
            [userprofilename]
            aws_access_key_id =  AKIA****************
            aws_secret_access_key = gMFRKAaeFM0*****************************
            mfa_serial = arn:aws:iam::<AccountID>:mfa/<IAM_user_name>
            region = ap-southeast-2

        Notice that the user profile contains `mfa_serial`.  This is not standard aws config however it is supported by GAC.  If it is available, GAC will use it to get temporary session tokens by requesting your associated MFA code.  This is useful for when your user has higher account priviledges that require MFA auth.


        Returns: commands to copy to your shell terminal

        -------------------------- ADVANCED USAGE --------------------------------------

        Consider appending the following to your `.bashrc` to have the environment variables set automatically:

        \b
            gac ()
            {
                for i in "$( get-aws-creds "$@" --quiet)";
                do
                    eval "${i}";
                done
            }


        Or alternatively, send the variables to a file and source them from your profile so that all terminal sessions receive the same temporary credentials:

            \b
            $ get-aws-creds --quiet > /tmp/awscreds
            $ source /tmp/awscreds
    """

    # Issue #15 - We may be trying to assume another role from a
    # shell that has previously had its temporary variables populated
    if 'AWS_SECRET_ACCESS_KEY' in os.environ:
        del os.environ['AWS_SECRET_ACCESS_KEY']
    if 'AWS_ACCESS_KEY_ID' in os.environ:
        del os.environ['AWS_ACCESS_KEY_ID']

    try:
        session = Session(profile=profile)
        creds = session.get_session_creds()
    except Exception as e:
        print("{}: {}".format(type(e).__name__, e))
        sys.exit(1)

    if not quiet:
        print("Keys and token for profile: '{profile}'".format(profile=profile))
        print("Paste the following into your shell:\n")

    print("export AWS_ACCESS_KEY_ID={}".format(creds.access_key))
    print("export AWS_SECRET_ACCESS_KEY={}".format(creds.secret_key))
    # non-temporary keys are returned by the boto3.Session if the
    # user is not required to use MFA token.  Printing out a null
    # field for this will cause malformed errors when using.
    if creds.token:
        print("export AWS_SESSION_TOKEN={}".format(creds.token))



class Singleton(type):
    """ Use as a metaclass. E.g:
            >>> class PeggedTime(metaclass=Singleton):
            ...     def __init__(self):
            ...         self.dt_now = dt.datetime.now()

        This class returns the same datetime no matter how many instances are created.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# flake8: noqa F401         - import not used error
from .table import add_key, append_to_list_field, add_new_map_field, clear_table, full_scan
from .aux import json_loads, json_dumps, DynamoEncoder
from .batch import batch_write, batch_delete, load_scan_dump, load_json_dump, get_ddb_table

name = 'dynamodb'
